import bpy
import bpy.types
import json
import os
import mathutils
from bpy.utils import previews
from bpy.app.handlers import persistent
from .FP_Utilities import wrap_label_text, debug_print, are_selected_meshes_valid, set_drawing_objects, create_fluids_collection, set_frame_driver


# Path and file definitions
script_directory = os.path.dirname(os.path.realpath(__file__))
image_directory = os.path.join(script_directory, "Pics")
if bpy.app.version >= (4, 2, 0):
    library_file_path = os.path.join(script_directory, "Library42", "Fluid_Painter_Library42.blend")
    json_file_path = os.path.join(script_directory, "Library42", "Fluid_Painter_Library42.json")
else:
    library_file_path = os.path.join(script_directory, "Library", "Fluid_Painter_Library.blend")
    json_file_path = os.path.join(script_directory, "Library", "Fluid_Painter_Library.json")

def reload_data():
    data.clear()
    update_data(json_file_path)
    

def update_data(file_path):
    # Load JSON file
    with open(file_path, 'r') as f:
        additional_data = json.load(f)
    
    # Merge the dictionaries
    data.update(additional_data)
    print(data)


data = {}
update_data(json_file_path)  # Load the main JSON file
suffix = ""

def set_addon_paths(other_addon_name, library_folder, library_file, image_dir):
    global additional_library_file_path
    global additional_image_directory
    global suffix
    
    suffix = other_addon_name.replace(__package__, "")
    print(suffix)

    addon_module = __import__(other_addon_name)
    additional_script_directory = os.path.dirname(os.path.realpath(addon_module.__file__))
    additional_library_file_path = os.path.join(additional_script_directory, library_folder, library_file)
    additional_image_directory = os.path.join(additional_script_directory, image_dir)
    print("addon path defined")

         
global pcoll
pcoll = None


def update_category(self, context):
    global pcoll
    category = context.scene.fluid_painter.category
    if category == "All":
        all_fluids = []
        for fluids in data.values():
            all_fluids.extend(fluids)
        context.scene.fluid_painter.fluid = all_fluids[0] if all_fluids else ''
    else:
        if data.get(category):
            context.scene.fluid_painter.fluid = data[category][0]
        else:
            context.scene.fluid_painter.fluid = ''

    update_image_preview(context)
    
    # print the initial selected fluid
    selected_fluid = context.scene.fluid_painter.fluid
    debug_print(f"Initial Selected Fluid: {selected_fluid}")


def update_fluid(self, context):
    global pcoll
    pcoll = update_image_preview(context)
    selected_fluid = context.scene.fluid_painter.fluid
    debug_print(f"Selected Fluid: {selected_fluid}")


def update_image_preview(context):
    global pcoll
    if hasattr(context, "scene"):
        debug_print("Context has attribute 'scene'")
        scene = context.scene
    elif hasattr(context, "view_layer"):
        debug_print("Context has attribute 'view_layer'")
        scene = context.view_layer.scene
    else:
        debug_print("Context has neither 'scene' nor 'view_layer'")
        return pcoll  # return current pcoll even if context doesn't have 'scene' or 'view_layer'

    fluid_name = scene.fluid_painter.fluid
    fluid_name_no_spaces = fluid_name.replace(" ", "")
    image_filename = f"Fluid_{fluid_name_no_spaces}_Image.png"

    if fluid_name in data.get(suffix, []):
        image_path = os.path.join(additional_image_directory, image_filename)
    else:
        image_path = os.path.join(image_directory, image_filename)

    # Skip creating image preview if fluid name is empty
    if not fluid_name:
        debug_print("Fluid name is empty")
        return pcoll  # return current pcoll even if fluid name is empty

    # Debug message for fluid name
    debug_print(f"Fluid name is {fluid_name}")

    # Create an image preview
    if pcoll:
        previews.remove(pcoll)
    pcoll = previews.new()
    try:
        preview_image = pcoll.load(fluid_name, image_path, 'IMAGE')
        debug_print(f"Image preview created for fluid: {fluid_name}")
    except Exception as e:
        debug_print(f"Failed to create image preview for fluid: {fluid_name}. Exception: {e}")

    return pcoll  # return current pcoll regardless of whether image preview creation was successful


@persistent
def update_image_preview_undo(scene):
    context = bpy.context
    global pcoll
    pcoll = update_image_preview(context)
    debug_print("Undo called")


def get_category_items(self, context):
    return [("All", "All", "All")] + [(i, i, i) for i in data.keys()]


def get_fluid_items(self, context):
    category = context.scene.fluid_painter.category
    if category == "All":
        all_fluids = []
        for fluids in data.values():
            all_fluids.extend(fluids)
        return [(i, i, i) for i in all_fluids]
    else:
        return [(i, i, i) for i in data.get(category, [])]


def update_fluidproperties():
    scene = bpy.context.scene
    if hasattr(scene, "fluid_painter") and isinstance(scene.fluid_painter, bpy.types.PropertyGroup):
        scene.fluid_painter.category = "All"
        scene.fluid_painter.fluid = 'Default'


class FluidPainterProperties(bpy.types.PropertyGroup):
    category: bpy.props.EnumProperty(
        items=get_category_items,
        update=update_category
    )
    fluid: bpy.props.EnumProperty(
        items=get_fluid_items,
        update=update_fluid
    )
    selected_fluid: bpy.props.StringProperty()


is_executing_add_from_panel = False  # Global lock variable

class AppendOperator(bpy.types.Operator):
    bl_idname = "object.append_fluid"
    bl_label = "Add Fluid"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        # Access preferences
        addon_preferences = context.preferences.addons[__package__].preferences
        fluid_is_topping = addon_preferences.fluid_is_topping
        
        # Check if in edit mode
        if context.object and context.object.mode == 'EDIT':
            return False

        # Check conditions for returning True
        selected_fluid = context.scene.fluid_painter.fluid
        selected_objects = bpy.context.selected_objects

        if ("+" in selected_fluid or fluid_is_topping) and are_selected_meshes_valid():
            return True

        if "+" not in selected_fluid and not fluid_is_topping:
            return True

        # Default case
        return False

    def execute(self, context):
        addon_preferences = context.preferences.addons[__package__].preferences
        edit_on_add = addon_preferences.edit_on_add
        fluid_scale_factor = addon_preferences.fluid_scale
        fluid_is_topping = addon_preferences.fluid_is_topping
        parent_fluid = addon_preferences.parent_fluid
        spline_resolution = addon_preferences.spline_resolution
    
        # Access the global variables
        global is_executing_add_from_panel
        is_executing_add_from_panel = True  # Set the lock variable to True
        
        # Get the selected fluid name from the drop down list
        fluid_name = context.scene.fluid_painter.fluid
        
        # Get the selected objects
        fluid_objects = bpy.context.selected_objects
        
        # Deselect all objects
        bpy.ops.object.select_all(action='DESELECT')
        
        if fluid_name in data.get(suffix, []):
            library_path = additional_library_file_path
        else:
            library_path = library_file_path
        
        # Append the selected fluid from the file
        with bpy.data.libraries.load(library_path) as (data_from, data_to):
            data_to.objects = [name for name in data_from.objects if name == fluid_name]

        # Get the fluid collection
        fluids_collection = create_fluids_collection()
        
        # Link or append the object to the current scene
        for obj in data_to.objects:
            if obj is not None:
                fluids_collection.objects.link(obj)
                
        # Set the active object to the last appended object
        fluid = bpy.context.collection.objects[-1]
        fluid.select_set(True)
        bpy.context.view_layer.objects.active = fluid

        fluid["fluid_init"] = 1
            
        # Get the curve data, delete all points and update the curve object to reflect the deletion
        curve = fluid.data
        curve.splines.clear()
        
        # Check if the fluid is a Fluid+
        if "fluid_plus" in fluid:
            set_drawing_objects(fluid, fluid_objects)
        
        # Set Parent
        if parent_fluid and len(fluid_objects) == 1:
            
            parent_object = fluid_objects[0]
            
            # Store the child's matrix world
            child_matrix_world = fluid.matrix_world.copy()

            # Set the child's parent to the parent object
            fluid.parent = parent_object

            # Reset the child's matrix world to its original position
            fluid.matrix_world = child_matrix_world

            # Update the matrix_parent_inverse to compensate for the parent's transformation
            inverse_matrix = mathutils.Matrix(fluid.matrix_world).inverted()
            fluid.matrix_parent_inverse = parent_object.matrix_world @ inverse_matrix
                  
        else:
            fluid.location = (0, 0, 0)    
        
        # Set the fluid_scale in the geometry node modifier
        fluid.modifiers["FluidPainterGN"]["Input_54"] = fluid_scale_factor
        
        # Set Loop Start and End Frame
        set_frame_driver(fluid.modifiers["FluidPainterGN"], "Input_77", "frame_start")
        set_frame_driver(fluid.modifiers["FluidPainterGN"], "Input_76", "frame_end")
        
        # Set Streaming Start and End Frame
        fluid.modifiers["FluidPainterGN"]["Input_79"] = bpy.context.scene.frame_start
        fluid.modifiers["FluidPainterGN"]["Input_80"] = bpy.context.scene.frame_end
        
        # Set the spline_resolution
        fluid.data.resolution_u = spline_resolution
        
        # Remove as asset
        bpy.ops.asset.clear()
        
        if edit_on_add:
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.wm.tool_set_by_id(name='builtin.draw')
            bpy.context.scene.tool_settings.curve_paint_settings.depth_mode = 'SURFACE'
        
        # Set the lock variable to False
        is_executing_add_from_panel = False 
            
        debug_print("Fluid added from Panel")
        
        return {'FINISHED'}
                    

def is_asset_browser_open():
    """Checks if the asset browser is open."""
    if bpy.context.screen:
        for area in bpy.context.screen.areas:
            if area.ui_type == 'ASSETS':
                return True
    return False

# Store the selection outside the function
selected_mesh = []
previous_objects = set()
process_fluid = False
depsgraph_update = 0
fluid = None
fluid_objects = []

@persistent       
def add_fluid_from_asset_browser(scene):

    addon_preferences = bpy.context.preferences.addons[__package__].preferences
    add_asset_browser = addon_preferences.add_asset_browser

    # Check if the asset browser is open or "add_asset_browser" is set to True
    if not is_asset_browser_open() or not add_asset_browser:
        return  

    global depsgraph_update
    depsgraph_update += 1
    debug_print("Depsgraph Update: ", depsgraph_update)
    
    edit_on_add = addon_preferences.edit_on_add
    fluid_scale_factor = addon_preferences.fluid_scale
    fluid_is_topping = addon_preferences.fluid_is_topping
    parent_fluid = addon_preferences.parent_fluid
    spline_resolution = addon_preferences.spline_resolution
    
    global is_executing_add_from_panel, selected_mesh, previous_objects, process_fluid, fluid, fluid_objects

    if not is_executing_add_from_panel:

        if process_fluid:
            # Set Parent
            if parent_fluid and len(fluid_objects) == 1:
                
                parent_object = fluid_objects[0]
                
                # Store the child's matrix world
                child_matrix_world = fluid.matrix_world.copy()

                # Set the child's parent to the parent object
                fluid.parent = parent_object

                # Reset the child's matrix world to its original position
                fluid.matrix_world = child_matrix_world

                # Update the matrix_parent_inverse to compensate for the parent's transformation
                inverse_matrix = mathutils.Matrix(fluid.matrix_world).inverted()
                fluid.matrix_parent_inverse = parent_object.matrix_world @ inverse_matrix
                      
            else:
                fluid.location = (0, 0, 0)
                
            # Set Loop Start and End Frame
            set_frame_driver(fluid.modifiers["FluidPainterGN"], "Input_77", "frame_start")
            set_frame_driver(fluid.modifiers["FluidPainterGN"], "Input_76", "frame_end")
            
            # Set Streaming Start and End Frame
            fluid.modifiers["FluidPainterGN"]["Input_79"] = bpy.context.scene.frame_start
            fluid.modifiers["FluidPainterGN"]["Input_80"] = bpy.context.scene.frame_end
            
            # Reset variables
            process_fluid = False
            fluid = None
            fluid_objects = []
        
        # Get the current list of objects
        current_objects = set(bpy.data.objects)

        # Check if any new objects have been added since the last invocation
        new_objects = current_objects - previous_objects

        # Update the previous object list for the next invocation
        previous_objects = current_objects

        # Check the selected objects
        current_selection = [
            obj for obj in bpy.context.selected_objects
            if obj.type == 'MESH'
            or obj.type == 'CURVE' and (obj.get("fluid_init", False) or "fluid_init" not in obj)
            and obj.name in bpy.data.objects
        ]

        debug_print("Current Selection: ", current_selection)

        curve_selected = None

        if new_objects:
            # At least one new object has been added
            for obj in new_objects:
                if "fluid_init" in obj and not obj["fluid_init"]:
                    curve_selected = obj
                    break  # Exit the loop after finding the first suitable object

        # Update selected_mesh if non-curve objects are selected or no objects are selected
        if current_selection or not curve_selected:
            selected_mesh = current_selection
                                
        if curve_selected and process_fluid == False:
            
            curve_selected["fluid_init"] = True
            debug_print("Curve: ", curve_selected.name)
            debug_print("Selected mesh: ", selected_mesh)
            
            # Get the selected objects
            fluid_objects = selected_mesh
            # Assuming only one curve is selected      
            fluid = curve_selected  
        
            if "fluid_plus" in fluid and fluid_objects == []:
                bpy.data.objects.remove(fluid, do_unlink=True)
                display_popup("Select one or more objects before adding a Fluid+ preset")
                return
            else:

                # Get the fluid collection
                fluids_collection = create_fluids_collection()

                # Deselect all objects
                bpy.ops.object.select_all(action='DESELECT')
                
                fluid.select_set(True)
                bpy.context.view_layer.objects.active = fluid
                    
                # Get the curve data, delete all points and update the curve object to reflect the deletion
                curve = fluid.data
                curve.splines.clear()
                
                # Check if the fluid is a Fluid+
                if "fluid_plus" in fluid:
                    set_drawing_objects(fluid, fluid_objects)
                
                # Set the fluid_scale in the geometry node modifier
                fluid.modifiers["FluidPainterGN"]["Input_54"] = fluid_scale_factor
                
                # Set the spline_resolution
                fluid.data.resolution_u = spline_resolution
                
                if edit_on_add:
                    bpy.ops.object.mode_set(mode='EDIT')
                    bpy.ops.wm.tool_set_by_id(name='builtin.draw')
                    bpy.context.scene.tool_settings.curve_paint_settings.depth_mode = 'SURFACE'
                   
                bpy.context.view_layer.objects.active = fluid
                
                debug_print("Fluid added from Asset Browser")
                
                process_fluid = True
            
            # Reset the variables after executing the rest of the code
            curve_selected = None
            selected_mesh = []


def display_popup_callback(self, context, message):
    layout = self.layout
    layout.label(text=message)


def display_popup(message):
    bpy.context.window_manager.popup_menu(lambda self, context: display_popup_callback(self, context, message), title="Warning!", icon='INFO')
    
    
# Define the panel class for Fluid Append
class FluidPainterAppendPanel(bpy.types.Panel):
    bl_label = "Fluid Add"
    bl_idname = "OBJECT_PT_fluid_add"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Fluid Painter'
    
    @classmethod
    def poll(cls, context):
        addon_preferences = context.preferences.addons[__package__].preferences
        add_fluid_panel = addon_preferences.add_fluid_panel

        # Enable the panel only if the add_fluid_panel flag is True
        return add_fluid_panel
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        fluid_painter = scene.fluid_painter
        selected_fluid = context.scene.fluid_painter.fluid
        
        # Get the width of the panel region
        panel_width = context.region.width
        
        # Fluid Selection
        layout.label(text="Fluid Preset:")
        if pcoll and fluid_painter.fluid in pcoll:
            layout.template_icon(pcoll[fluid_painter.fluid].icon_id, scale=10)

        layout.prop(fluid_painter, "category", text='Category')
        layout.prop(fluid_painter, "fluid", text='Fluid')
        
        addon_preferences = context.preferences.addons[__package__].preferences
        
        layout.label(text="Fluid Initial Settings:")
        
        # Access the fluid_scale property directly from addon_preferences
        layout.prop(addon_preferences, "fluid_scale", text="Fluid Scale")
        
        # Access the spline_resolution property directly from addon_preferences
        layout.prop(addon_preferences, "spline_resolution", text="Fluid Spline Resolution")
        
        #layout.prop(addon_prefs, "fluid_is_topping")
        layout.prop(addon_preferences, "parent_fluid", text="Parent Fluid to selected object")

        # Access the fluid_is_topping property directly from addon_preferences
        fluid_is_topping = addon_preferences.fluid_is_topping
        parent_fluid = addon_preferences.parent_fluid

        box = layout.box()
        
        box.operator("object.append_fluid", text="Add Fluid")
        
        if "+" in selected_fluid or fluid_is_topping:
            
            # Check if there is just one selected mesh object
            if len(bpy.context.selected_objects) == 1 and are_selected_meshes_valid():
                box.label(text="Drawing Object: " + bpy.context.selected_objects[0].name)

            # Check if there are multiple selected mesh objects
            elif len(bpy.context.selected_objects) >= 2 and are_selected_meshes_valid():
                mesh_names = [obj.name for obj in bpy.context.selected_objects]
                wrapped_lines = wrap_label_text("New Drawing Collection: " + selected_fluid + " Collection", panel_width)
                for line in wrapped_lines:
                    box.label(text=line)
            else:
                if fluid_is_topping:
                    wrapped_lines = wrap_label_text("Warning! Select one or more objects before adding a Fluid as Topping.", panel_width)
                    for line in wrapped_lines:
                        box.label(text=line)
                
                if "+" in selected_fluid:
                    wrapped_lines = wrap_label_text("Warning! Select one or more objects before adding a Fluid+ Preset.", panel_width)
                    for line in wrapped_lines:
                        box.label(text=line)
                    
        if parent_fluid:
            if len(bpy.context.selected_objects) == 1 and are_selected_meshes_valid():
                box.label(text="Parent Object: " + bpy.context.selected_objects[0].name)
        
            elif len(bpy.context.selected_objects) == 0:
                wrapped_lines = wrap_label_text("Warning! No Object selected, parent won't be set.", panel_width)
                for line in wrapped_lines:
                    box.label(text=line)
                        
            elif len(bpy.context.selected_objects) > 1:
                wrapped_lines = wrap_label_text("Warning! More than one object selected, parent won't be set.", panel_width)
                for line in wrapped_lines:
                    box.label(text=line)