import bpy
from .FP_Utilities import is_fluid_curve, is_fluid_curve_uptodate, set_drawing_objects, create_fluids_collection, wrap_label_text, set_frame_driver


# Define the operator to start drawing
class StartDrawingOperator(bpy.types.Operator):
    bl_idname = "object.start_drawing"
    bl_label = "Start Drawing (Tab)"

    @classmethod
    def poll(cls, context):
        return context.object.type == 'CURVE'

    def execute(self, context):
        bpy.ops.object.mode_set(mode='EDIT')
        # Set the draw tool
        bpy.ops.wm.tool_set_by_id(name='builtin.draw')
        bpy.context.scene.tool_settings.curve_paint_settings.depth_mode = 'SURFACE'
        return {'FINISHED'}


# Define the operator to stop drawing
class StopDrawingOperator(bpy.types.Operator):
    bl_idname = "object.stop_drawing"
    bl_label = "Stop Drawing (Tab)"

    @classmethod
    def poll(cls, context):
        return context.object.type == 'CURVE'
    
    def execute(self, context):
        bpy.ops.object.mode_set(mode='OBJECT')
        return {'FINISHED'}


class SetDrawingObjectsOperator(bpy.types.Operator):
    bl_idname = "object.set_drawing_objects"
    bl_label = "Set Drawing Objects"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        # Disabled if not in Object mode or there are less than two objects selected
        if context.mode != 'OBJECT' or len(context.selected_objects) < 2:
            return False

        selected_objects = context.selected_objects

        mesh_selected = False
        curve_selected = False

        for obj in selected_objects:
            if obj.type == 'MESH' or (obj.type == 'CURVE' and (obj.get("fluid_init", False) or "fluid_init" not in obj)):
                mesh_selected = True

        active_object = context.active_object
        if active_object and active_object.type == 'CURVE' and active_object.get("fluid_plus", False):
            curve_selected = True

        return mesh_selected and curve_selected


    def execute(self, context):
        selected_objects = bpy.context.selected_objects
        fluid = None
        
        for obj in selected_objects:
            if obj.type == "CURVE" and obj.get("fluid_plus", False):
                fluid = obj
                break

        # Deselect the fluid object
        fluid.select_set(False)
        
        set_drawing_objects(fluid, bpy.context.selected_objects)
        
        # Deselect all objects
        bpy.ops.object.select_all(action='DESELECT')

        # Set fluid as the active object and select it
        fluid.select_set(True)
        bpy.context.view_layer.objects.active = fluid
            
        return {'FINISHED'}


class SetFrameDriverOperator(bpy.types.Operator):
    bl_idname = "object.set_frame_driver"
    bl_label = "Set Frame Driver"

    def execute(self, context):
        # Access the active object
        active_obj = context.active_object

        # Access the modifier by name
        modifier = active_obj.modifiers.get('FluidPainterGN')

        # Check if modifier exists and "Input_75" is 1
        if modifier and modifier["Input_75"] == 1:
            # Call set_frame_driver function with provided arguments
            set_frame_driver(modifier, "Input_77", "frame_start")
            set_frame_driver(modifier, "Input_76", "frame_end")
        
        return {'FINISHED'}


class SetSingleStreamOperator(bpy.types.Operator):
    bl_idname = "object.set_single_stream"
    bl_label = "Set Single Drop Stream"

    def execute(self, context):
        
        # Access the active object
        active_obj = context.active_object

        # Access the modifier by name
        modifier = active_obj.modifiers.get('FluidPainterGN')

        # Check if modifier exists and "Input_78" is bigger than 1
        if modifier and modifier["Input_78"] >= 1:
            modifier["Input_83"] = not modifier["Input_83"]
            
        active_obj.update_tag()

        return {'FINISHED'}


# Define the panel class for Fluid Editing
class FluidPainterEditingPanel:
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "Fluid Painter"


class FluidPainterEditingPanelMain(FluidPainterEditingPanel, bpy.types.Panel):
    bl_label = "Fluid Editing"
    bl_idname = "OBJECT_PT_fluid_editing"
    
    def draw(self, context):
        # Get the width of the panel region
        panel_width = context.region.width
        has_drawing_objects = False
        layout = self.layout
        
        if is_fluid_curve(context.object) and "fluid_plus" in context.object:
            target_curve = context.object
            if target_curve.modifiers["FluidPainterGN"]["Input_41"] == None and target_curve.modifiers["FluidPainterGN"]["Input_36"] == None:
                has_drawing_objects = False
            elif target_curve.modifiers["FluidPainterGN"]["Input_41"] != None or target_curve.modifiers["FluidPainterGN"]["Input_36"] != None:
                has_drawing_objects = True
        
        box = layout.box()
        
        if len(context.selected_objects) == 1:
            # Get the selected object
            selected_object = context.selected_objects[0]
        
            if is_fluid_curve(selected_object):
                # Check if in edit mode
                if context.object and context.object.mode == 'EDIT':
                    box.operator("object.stop_drawing")
                else:
                    box.operator("object.start_drawing")
                
                if not is_fluid_curve_uptodate(selected_object):
                    wrapped_lines = wrap_label_text("Fluid is outdated, it is recommended to update it using the Fluid Update Panel", panel_width)
                    for line in wrapped_lines:
                        box.label(text=line)

            elif any(mod.type == 'NODES' and mod.node_group.name.startswith("Fluid Painter") for mod in context.active_object.modifiers):
                wrapped_lines = wrap_label_text("Fluid is outdated, update it from the Fluid Update Panel", panel_width)
                for line in wrapped_lines:
                    box.label(text=line)
            
            else:
                wrapped_lines = wrap_label_text("Select a Fluid Curve", panel_width)
                for line in wrapped_lines:
                    box.label(text=line)
            
            if "fluid_plus" in selected_object and has_drawing_objects == False:
                wrapped_lines = wrap_label_text("Warning! Fluid needs one or more drawing objects. Select one or more objects, then the curve and click on Set Drawing Objects", panel_width)
                for line in wrapped_lines:
                    box.label(text=line)
                
        else:
            wrapped_lines = wrap_label_text("Select a Fluid Curve", panel_width)
            for line in wrapped_lines:
                box.label(text=line)
                
        # Drawing Object Logic
        if is_fluid_curve(context.object) and "fluid_plus" in context.object:
            target_curve = context.object
            if target_curve.modifiers["FluidPainterGN"]["Input_41"] == None and target_curve.modifiers["FluidPainterGN"]["Input_36"] == None:
                box.operator("object.set_drawing_objects", text="Set Drawing Objects")
                has_drawing_objects = False    
            elif target_curve.modifiers["FluidPainterGN"]["Input_41"] != None or target_curve.modifiers["FluidPainterGN"]["Input_36"] != None:
                box.operator("object.set_drawing_objects", text="Update Drawing Objects")
                has_drawing_objects = True
        

class FluidPainterEditingPanelProperties(FluidPainterEditingPanel, bpy.types.Panel):
    bl_parent_id = "OBJECT_PT_fluid_editing"
    bl_idname = "OBJECT_PT_fluid_properties"
    bl_label = "Properties"

    @classmethod
    def poll(cls, context):
        if len(context.selected_objects) == 1:
            selected_object = context.selected_objects[0]
            return is_fluid_curve(selected_object)
        return False

    def draw(self, context):
        layout = self.layout
       
        # Curve Spline Resolution
        layout.prop(context.active_object.data, "resolution_u", text="Fluid Spline Resolution")
        
        obj = context.active_object
        modifier = context.active_object.modifiers.get('FluidPainterGN')
        
        # Check if '+' is present in object name
        show_plus_inputs = "fluid_plus" in context.active_object

        if modifier and modifier.node_group.name.startswith("Fluid Painter") and modifier.node_group:
            node_group = modifier.node_group
            
            if bpy.app.version >= (4, 0, 0):
                inputs = node_group.interface
                fluid_inputs = []
                for input_name in inputs.items_tree:
                    if input_name.name.startswith('Fluid'):
                        fluid_inputs.append(input_name)
                pass
            else:
                inputs = node_group.inputs
                fluid_inputs = []
                for input_name in inputs:
                    if input_name.name.startswith('Fluid'):
                        fluid_inputs.append(input_name)
                pass

            for input_name in fluid_inputs:
                if (show_plus_inputs and '+' in input_name.name and input_name.identifier == "Input_41"):
                    layout.prop_search(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        search_data=bpy.data,
                        search_property="objects",
                        text=(input_name.name)
                    )
                    
                elif (show_plus_inputs and '+' in input_name.name and input_name.identifier == "Input_36"):
                    layout.prop_search(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        search_data=bpy.data,
                        search_property="collections",
                        text=(input_name.name)
                    )
                    
                elif (show_plus_inputs or '+' not in input_name.name):
                    layout.prop(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        text=(input_name.name)
                    )
                    

class FluidPainterEditingPanelMaterial(FluidPainterEditingPanel, bpy.types.Panel):
    bl_parent_id = "OBJECT_PT_fluid_editing"
    bl_idname = "OBJECT_PT_fluid_material"
    bl_label = "Material"
    
    @classmethod
    def poll(cls, context):
        if len(context.selected_objects) == 1:
            selected_object = context.selected_objects[0]
            return is_fluid_curve(selected_object)
        return False
        
    def draw(self, context):
        layout = self.layout
        
        modifier = context.active_object.modifiers.get('FluidPainterGN')
        
        # Check if '+' is present in object name
        show_plus_inputs = "fluid_plus" in context.active_object
        
        if modifier and modifier.node_group.name.startswith("Fluid Painter") and modifier.node_group:
            node_group = modifier.node_group
            
            if bpy.app.version >= (4, 0, 0):
                inputs = node_group.interface
                fluid_inputs = []
                for input_name in inputs.items_tree:
                    if input_name.name.startswith('Material'):
                        fluid_inputs.append(input_name)
                pass
            else:
                inputs = node_group.inputs
                fluid_inputs = []
                for input_name in inputs:
                    if input_name.name.startswith('Material'):
                        fluid_inputs.append(input_name)
                pass

            for input_name in fluid_inputs:
                if input_name.identifier == "Input_3":
                    layout.prop_search(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        search_data=bpy.data,
                        search_property="materials",
                        text=(input_name.name)
                    )
                    
                elif (show_plus_inputs or '+' not in input_name.name) and \
                    (input_name.identifier != "Input_70" or modifier.get("Output_71_attribute_name") == "rot") and \
                    (input_name.identifier != "Input_74" or modifier.get("Output_73_attribute_name") == "base_color") and \
                    (input_name.identifier != "Input_85" or modifier.get("Output_84_attribute_name") == "shadow") and \
                    (input_name.identifier != "Input_53" or modifier.get("Output_52_attribute_name") == "uv_scale"):
                    
                    layout.prop(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        text=(input_name.name)
                    )
        
        

class FluidPainterEditingPanelDrop(FluidPainterEditingPanel, bpy.types.Panel):
    bl_parent_id = "OBJECT_PT_fluid_editing"
    bl_idname = "OBJECT_PT_fluid_drop"
    bl_label = "Drop"
    bl_options = {"DEFAULT_CLOSED"}
    
    @classmethod
    def poll(cls, context):
        if len(context.selected_objects) == 1:
            selected_object = context.selected_objects[0]
            return is_fluid_curve(selected_object)
        return False

    def draw(self, context):
        layout = self.layout
        
        modifier = context.active_object.modifiers.get('FluidPainterGN')

        if modifier and modifier.node_group.name.startswith("Fluid Painter") and modifier.node_group:
            node_group = modifier.node_group
            
            if bpy.app.version >= (4, 0, 0):
                inputs = node_group.interface
                fluid_inputs = []
                for input_name in inputs.items_tree:
                    if input_name.name.startswith('Drop'):
                        fluid_inputs.append(input_name)
                pass
            else:
                inputs = node_group.inputs
                fluid_inputs = []
                for input_name in inputs:
                    if input_name.name.startswith('Drop'):
                        fluid_inputs.append(input_name)
                pass
            
            for input_name in fluid_inputs:
                if input_name.identifier == "Input_75":
                    box = layout.box()
                    box.prop(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        text=(input_name.name)
                        )
                    if modifier["Input_75"] == 1:
                        row = box.row(align=True)
                        col = row.column()
                        col.operator("object.set_frame_driver", text="", icon="FILE_REFRESH")
                        
                        col = row.column()
                        col.prop(
                            data=modifier,
                            property='["Input_77"]',
                            text="Start"
                        )

                        col = row.column()
                        col.prop(
                            data=modifier,
                            property='["Input_76"]',
                            text="End",
                        )
                
                elif input_name.identifier == "Input_78":
                    box = layout.box()
                    box.prop(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        text=(input_name.name)
                        )
                    if modifier["Input_78"] >= 1:
                        row = box.row(align=True)
                        
                        col = row.column()
                        col.operator("object.set_single_stream", text="", icon="FORWARD")
                        
                        col = row.column()
                        col.prop(
                            data=modifier,
                            property='["Input_79"]',
                            text="Start"
                        )
                        
                        if modifier["Input_83"] == 0:
                            col = row.column()
                            col.prop(
                                data=modifier,
                                property='["Input_80"]',
                                text="End"
                            )

                elif input_name.identifier != "Input_76" and input_name.identifier != "Input_77" and input_name.identifier != "Input_79" and input_name.identifier != "Input_80":
                    layout.prop(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        text=(input_name.name),
                    )
            
            
class FluidPainterEditingPanelLayerA(FluidPainterEditingPanel, bpy.types.Panel):
    bl_parent_id = "OBJECT_PT_fluid_editing"
    bl_idname = "OBJECT_PT_fluid_layerA"
    bl_label = "Layer A"
    bl_options = {"DEFAULT_CLOSED"}
    
    @classmethod
    def poll(cls, context):
        if len(context.selected_objects) == 1:
            selected_object = context.selected_objects[0]
            return is_fluid_curve(selected_object)
        return False

    def draw(self, context):
        layout = self.layout
        
        modifier = context.active_object.modifiers.get('FluidPainterGN')

        if modifier and modifier.node_group.name.startswith("Fluid Painter") and modifier.node_group:
            node_group = modifier.node_group
            
            if bpy.app.version >= (4, 0, 0):
                inputs = node_group.interface
                fluid_inputs = []
                for input_name in inputs.items_tree:
                    if input_name.name.startswith('Layer A'):
                        fluid_inputs.append(input_name)
                pass
            else:
                inputs = node_group.inputs
                fluid_inputs = []
                for input_name in inputs:
                    if input_name.name.startswith('Layer A'):
                        fluid_inputs.append(input_name)
                pass

            for input_name in fluid_inputs:
                if input_name.identifier == "Input_25":
                    layout.prop_search(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        search_data=bpy.data,
                        search_property="materials",
                        text=(input_name.name)
                    )
                else:
                    layout.prop(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        text=(input_name.name)
                    )
 
 
class FluidPainterEditingPanelLayerB(FluidPainterEditingPanel, bpy.types.Panel):
    bl_parent_id = "OBJECT_PT_fluid_editing"
    bl_idname = "OBJECT_PT_layerB"
    bl_label = "Layer B"
    bl_options = {"DEFAULT_CLOSED"}
    
    @classmethod
    def poll(cls, context):
        if len(context.selected_objects) == 1:
            selected_object = context.selected_objects[0]
            return is_fluid_curve(selected_object)
        return False

    def draw(self, context):
        layout = self.layout
        
        modifier = context.active_object.modifiers.get('FluidPainterGN')

        if modifier and modifier.node_group.name.startswith("Fluid Painter") and modifier.node_group:
            node_group = modifier.node_group
            
            if bpy.app.version >= (4, 0, 0):
                inputs = node_group.interface
                fluid_inputs = []
                for input_name in inputs.items_tree:
                    if input_name.name.startswith('Layer B'):
                        fluid_inputs.append(input_name)
                pass
            else:
                inputs = node_group.inputs
                fluid_inputs = []
                for input_name in inputs:
                    if input_name.name.startswith('Layer B'):
                        fluid_inputs.append(input_name)
                pass

            for input_name in fluid_inputs:
                if input_name.identifier == "Input_26":
                    layout.prop_search(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        search_data=bpy.data,
                        search_property="materials",
                        text=(input_name.name)
                    )
                else:
                    layout.prop(
                        data=modifier,
                        property='["{}"]'.format(input_name.identifier),
                        text=(input_name.name)
                    )