import bpy
import os
import mathutils
from . import main_fluid_version
from .FP_Utilities import is_fluid_curve, is_fluid_curve_not_sel, wrap_label_text, create_fluids_collection, clear_tags, get_additional_library_file_path, debug_print, set_frame_driver
from .FP_Append import library_file_path


class ConvertToMeshOperator(bpy.types.Operator):
    bl_idname = "object.convert_to_mesh"
    bl_label = "Convert to Mesh"
    bl_options = {'REGISTER', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        if len(context.selected_objects) == 1:
            # Get the selected object
            selected_object = context.selected_objects[0]

        # Disabled if not in Object mode or the select object is not a fluid curve
        return context.mode == 'OBJECT' and is_fluid_curve(selected_object)
            
    def execute(self, context):
        if context.object:
            # Convert to Mesh the active object
            bpy.ops.object.convert(target='MESH')
            
            return {'FINISHED'}


class SetOriginOperator(bpy.types.Operator):
    bl_idname = "object.set_origin"
    bl_label = "Set Origin to Geometry"
    bl_options = {'REGISTER', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        # Check if Blender is in Object Mode
        if context.mode != 'OBJECT':
            return False

        # Check if there's exactly one selected object
        if len(context.selected_objects) != 1:
            return False

        # Get the selected object
        selected_object = context.selected_objects[0]

        # Check if the selected object is a fluid curve
        return is_fluid_curve(selected_object)
            
    def execute(self, context):
        if context.object:
            # Set the origin to geometry for the active object
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY', center='BOUNDS')            
            return {'FINISHED'}


def get_armature_and_pose_position(obj):
    if obj and obj.type == 'MESH':
        armature_modifier = next((mod for mod in obj.modifiers if mod.type == 'ARMATURE'), None)
        
        if armature_modifier:
            armature = armature_modifier.object
            if armature and armature.type == 'ARMATURE':
                return armature, armature.data.pose_position

    return None, None


class SwitchRestPose(bpy.types.Operator):
    bl_idname = "object.switch_rest_pose"
    bl_label = "Switch Rest/Pose"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        sticking_mesh = context.scene.sticking_mesh
        # Check if the armature from sticking_mesh is in pose mode
        armature, current_position = get_armature_and_pose_position(sticking_mesh)
        
        # Disable if there is no valid armature
        if not armature:
            return False

        # Enable the operator
        return True

    def execute(self, context):
        # Get the object from the scene property
        obj = context.scene.sticking_mesh
        armature, current_position = get_armature_and_pose_position(obj)

        if armature and current_position:
            if current_position == 'REST':
                armature.data.pose_position = 'POSE'
            else:
                armature.data.pose_position = 'REST'

        return {'FINISHED'}


# Operator to convert the selected curve to use the Fluid Painter node group
class ConvertToFluidOperator(bpy.types.Operator):
    bl_idname = "object.convert_to_fluid_operator"
    bl_label = "Convert to Fluid"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        # Check if there are any selected objects
        if not context.selected_objects:
            return False
    
        # Check if there's just one selected object
        if context.mode != 'OBJECT' and len(context.selected_objects) != 1:
            return False
        
        selected_object = context.selected_objects[0]

        # Check if the selected object is of type 'CURVE'
        if selected_object.type != 'CURVE':
            return False

        # Check if the selected object has a geometry node modifier with a node group name that starts with "fluid painter"
        has_fluid_painter_modifier = any(mod.type == 'NODES' and mod.node_group.name.startswith("Fluid Painter") for mod in selected_object.modifiers)
        
        # Return True only if the conditions are met
        if has_fluid_painter_modifier:
            return False

        return True

    def execute(self, context):
        addon_preferences = bpy.context.preferences.addons[__package__].preferences
        fluid_scale_factor = context.scene.fluid_scale
        
        # Get the active object (selected curve)
        selected_object = bpy.context.selected_objects[0]
        
        # Check conditions for returning True
        selected_fluid = context.scene.fluid_painter.fluid
        print("selected fluid is: "+ str(selected_fluid))
        
        # Define the list of library paths to check
        library_paths = [library_file_path]
        
        additional_library_file_path = get_additional_library_file_path()

        # Check if additional_library_file_path is not empty and add it to the list
        if additional_library_file_path:
            library_paths.append(additional_library_file_path)
            debug_print(f"Library Paths: {library_paths}")
        
        # Iterate through the library paths and try to import the material
        for library_path in library_paths:
            with bpy.data.libraries.load(library_path, link=False) as (data_from, data_to):
                # Check if the selected_fluid object is in the current library
                if selected_fluid in data_from.objects:
                    # If found, assign it to data_to.objects and break the loop
                    data_to.objects = [selected_fluid]
                    break

        # Initialize a variable to hold the last appended object
        temp_fluid = None
        
        # Append the object to the current scene
        for obj in data_to.objects:
            if obj is not None:
                # Link the object to the scene
                bpy.context.scene.collection.objects.link(obj)
                # Assign the last appended object to the variable
                temp_fluid = obj
                
        if temp_fluid is not None:
            temp_fluid["fluid_init"] = 1
            temp_fluid.select_set(True)
            selected_object.select_set(True)
            bpy.context.view_layer.objects.active = temp_fluid
            
            # Copy the FluidPainterGN modifier from the original to the copy
            bpy.ops.object.modifier_copy_to_selected(modifier="FluidPainterGN")
            
            # Transfer materials from temp_fluid to selected_object
            for slot in temp_fluid.material_slots:
                selected_object.data.materials.append(slot.material.copy())
        else:
            print("No object found in the library.")

        # Deselect original_obj
        temp_fluid.select_set(False)
        
        # Set the fluid_scale in the geometry node modifier
        selected_object.modifiers["FluidPainterGN"]["Input_54"] = fluid_scale_factor
        
        # Set End Frame
        modifier = selected_object.modifiers.get('FluidPainterGN')
        set_frame_driver(modifier, "Input_77", "frame_start")
        set_frame_driver(modifier, "Input_76", "frame_end")

        # Add "fluid_init" property and set to true
        selected_object["fluid_init"] = True

        # Check if the curve name contains a "+"
        if "+" in temp_fluid.name:
            # Add "fluid_plus" property and set to true
            selected_object["fluid_plus"] = True
            
        # Add version property
        selected_object["fluid_ver"] = main_fluid_version
        
        # Delete temp_fluid
        bpy.data.objects.remove(temp_fluid, do_unlink=True)

        # Update the copy object
        selected_object.update_tag()
        
        # Get the fluid collection
        fluids_collection = create_fluids_collection()
        
        # Unlink the object from its current collection (if it's linked to any)
        if selected_object.users_collection:
            current_collection = selected_object.users_collection[0]
            current_collection.objects.unlink(selected_object)
        
        fluids_collection.objects.link(selected_object)
        
        # Set the selected_object as the active object and select it
        bpy.context.view_layer.objects.active = selected_object
        selected_object.select_set(True)

        return {'FINISHED'}


def set_parent(child_obj, parent_obj):
    # Check if both objects exist
    if child_obj and parent_obj:
        # Select both objects
        bpy.context.view_layer.objects.active = parent_obj
        parent_obj.select_set(True)
        child_obj.select_set(True)

        # Set parenting using bpy.ops.object.parent_set
        bpy.ops.object.parent_set(type='OBJECT', keep_transform=True)

        # Clear the selection
        bpy.context.view_layer.objects.active = None
        bpy.ops.object.select_all(action='DESELECT')         


class VexterGroupsTransfer(bpy.types.Operator):
    bl_idname = "object.vertex_groups_transfer"
    bl_label = "Vertex Groups transfer"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        # Store commonly used attributes of the context
        mode = context.mode
        selected_objects = context.selected_objects
        sticking_mesh = context.scene.sticking_mesh
        
        # Check if sticking_mesh is None
        if sticking_mesh is None:
            return False
        
        # Check if the collection containing the stiching mesh is hidden
        if sticking_mesh.users_collection:
            collection = sticking_mesh.users_collection[0]  # Assuming the mesh is only in one collection
            view_layer = context.view_layer
            for layer_collection in view_layer.layer_collection.children:
                if layer_collection.collection == collection:
                    if layer_collection.exclude:
                        return False
                    break
        
        # Check if Blender is in Object Mode
        if mode != 'OBJECT':
            return False
    
        # Check if anything is selected
        if not selected_objects:
            return False
    
        # Check if all selected objects are valid fluids
        if not all(is_fluid_curve(obj) for obj in selected_objects):
            return False
            
        # Check if all selected objects are curves
        if not all(obj.type == 'CURVE' for obj in selected_objects):
            return False

        # Check if the sticking_mesh object is valid
        if not sticking_mesh or sticking_mesh.type != 'MESH':
            return False

        # Check if the armature from sticking_mesh is in pose mode
        armature, current_position = get_armature_and_pose_position(sticking_mesh)
        # Disable if there is no valid armature or if it is already in pose mode
        if not armature or current_position == 'POSE':
            return False

        # Enable the operator
        return True


    def execute(self, context):
    
        # Get the selected objects
        selected_objects = bpy.context.selected_objects

        for original_obj in selected_objects:
    
            # Set the spline_resolution
            original_obj.data.resolution_u = context.scene.spline_resolution

            # Create a copy of the original object
            original_name = original_obj.name
            copy_obj = original_obj.copy()
            copy_obj.data = original_obj.data.copy()
            # Link the copy_obj to the same collection as original_obj
            original_obj_collection = original_obj.users_collection[0]
            original_obj_collection.objects.link(copy_obj)
            
            # Hide the original object
            #original_obj.hide_set(True)
            original_obj.hide_viewport = True
            original_obj.hide_render = True

            # Make the copy the active object
            bpy.context.view_layer.objects.active = copy_obj
            
            bpy.ops.object.select_all(action='DESELECT')
            
            copy_obj.select_set(True)
            
            # Make the copy object visible
            copy_obj.hide_set(False)
            copy_obj.hide_render = False
            
            if context.scene.keep_fluid_gn:
                # Rename the cpy object with "_RIGGED_CURVE" suffix
                copy_obj.name = original_name + "_RIGGED_CURVE"
            
                # Remove the Geometry Nodes modifier from the copy if present
                for modifier in copy_obj.modifiers:
                    if modifier.type == 'NODES' and modifier.name == "FluidPainterGN":
                        copy_obj.modifiers.remove(modifier)  
            else:
                # Rename the cpy object with "_RIGGED_MESH" suffix
                copy_obj.name = original_name + "_RIGGED_MESH"
                clear_tags(copy_obj)

            # Convert the curve to mesh
            bpy.ops.object.convert(target='MESH')

            # Get the active object (now a mesh)
            bpy.ops.object.select_all(action='DESELECT')

            # Get the source object from the scene property
            source_obj = context.scene.sticking_mesh
            armature_obj = source_obj.find_armature()
            set_parent(copy_obj, armature_obj)

            # Select both objects
            source_obj.select_set(True)
            copy_obj.select_set(True)
            
            # Set the active object to the original object
            bpy.context.view_layer.objects.active = source_obj

            bpy.ops.object.data_transfer(
                use_freeze=False,
                data_type='VGROUP_WEIGHTS',
                use_create=True,
                vert_mapping='NEAREST', 
                use_auto_transform=False,
                use_object_transform=True,
                use_max_distance=False,
                ray_radius=0,
                layers_select_src='ALL',
                layers_select_dst='NAME',
                mix_mode='REPLACE',
                mix_factor=1
            )

            # Deselect both objects
            source_obj.select_set(False)

            # Make the copy the active object
            bpy.context.view_layer.objects.active = copy_obj

            # Add an Armature modifier with the armature of the source object
            armature_modifier = copy_obj.modifiers.new(name="Armature", type='ARMATURE')
            armature_modifier.object = source_obj.find_armature()  # Use the armature object of the source
            # Enable "Deform" and "Preserve Volume" options
            copy_obj.modifiers["Armature"].use_deform_preserve_volume = True

            # Check if "FluidPainterGN" Geometry Nodes modifier exists in the original object
            original_gn_modifier = original_obj.modifiers.get("FluidPainterGN")
            if original_gn_modifier and context.scene.keep_fluid_gn:
                # Make the original object active and the copy object selected
                original_obj.select_set(True)
                copy_obj.select_set(True)
                bpy.context.view_layer.objects.active = original_obj
                
                # Copy the FluidPainterGN modifier from the original to the copy
                bpy.ops.object.modifier_copy_to_selected(modifier="FluidPainterGN")
                
                # Deselect both objects
                original_obj.select_set(False)
                
                bpy.context.view_layer.objects.active = copy_obj
                
                # Set the "Is Mesh" flag to true
                copy_obj.modifiers["FluidPainterGN"]["Input_65"] = True
                            
                # Update the copy object
                copy_obj.update_tag()

            self.report({'INFO'}, "Data Transfer and Armature modifiers added to " + copy_obj.name)
        
        return {'FINISHED'}


class BindToMesh(bpy.types.Operator):
    bl_idname = "object.bind_to_mesh"
    bl_label = "Bind to Mesh"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        # Store commonly used attributes of the context
        mode = context.mode
        selected_objects = context.selected_objects
        sticking_mesh = context.scene.sticking_mesh
        
        # Check if sticking_mesh is None
        if sticking_mesh is None:
            return False
        
        # Check if the collection containing the stiching mesh is hidden
        if sticking_mesh.users_collection:
            collection = sticking_mesh.users_collection[0]  # Assuming the mesh is only in one collection
            view_layer = context.view_layer
            for layer_collection in view_layer.layer_collection.children:
                if layer_collection.collection == collection:
                    if layer_collection.exclude:
                        return False
                    break
        
        # Check if Blender is in Object Mode
        if mode != 'OBJECT':
            return False
        
        # Check if anything is selected
        if not selected_objects:
            return False
        
        # Check if all selected objects are valid fluids
        if not all(is_fluid_curve(obj) for obj in selected_objects):
            return False
            
        # Check if all selected objects are curves
        if not all(obj.type == 'CURVE' for obj in selected_objects):
            return False

        # Check if the sticking_mesh object is valid
        if not sticking_mesh or sticking_mesh.type != 'MESH':
            return False

        # Enable the operator
        return True

    def execute(self, context):

        # Get the selected objects
        selected_objects = bpy.context.selected_objects

        for original_obj in selected_objects:
    
            # Set the spline_resolution
            if context.scene.keep_fluid_gn:
                original_obj.data.resolution_u = context.scene.spline_resolution

            # Create a copy of the original object
            original_name = original_obj.name
            copy_obj = original_obj.copy()
            copy_obj.data = original_obj.data.copy()
            
            # Link the copy_obj to the same collection as original_obj
            original_obj_collection = original_obj.users_collection[0]
            original_obj_collection.objects.link(copy_obj)
          
            # Hide the original object
            #original_obj.hide_set(True)
            original_obj.hide_viewport = True
            original_obj.hide_render = True
            
            # Make the copy the active object
            bpy.context.view_layer.objects.active = copy_obj
            
            bpy.ops.object.select_all(action='DESELECT')
            
            copy_obj.select_set(True)
            
            # Make the copy object visible
            copy_obj.hide_set(False)
            copy_obj.hide_render = False
    
            if context.scene.keep_fluid_gn:
                # Rename the cpy object with "_RIGGED_CURVE" suffix
                copy_obj.name = original_name + ".BOUND_CURVE"
            
                # Remove the Geometry Nodes modifier from the copy if present
                for modifier in copy_obj.modifiers:
                    if modifier.type == 'NODES' and modifier.name == "FluidPainterGN":
                        copy_obj.modifiers.remove(modifier)  
            else:
                # Rename the cpy object with "_RIGGED_MESH" suffix
                copy_obj.name = original_name + ".BOUND_MESH"
                clear_tags(copy_obj)

            # Convert the curve to mesh
            bpy.ops.object.convert(target='MESH')

            # Deselect all
            bpy.ops.object.select_all(action='DESELECT')

            # Get the source object from the scene property
            binding_source = context.scene.sticking_mesh

            set_parent(copy_obj, binding_source)
            
            # Make the copy the active object
            bpy.context.view_layer.objects.active = copy_obj
            
            bpy.ops.object.select_all(action='DESELECT')
            
            copy_obj.select_set(True)

            # Check if there's any Subsurface or Multires Modifier
            subsurf_detected = False
            multires_detected = False
            
            subsurf_modifier = None
            multires_modifier = None
            
            for modifier in binding_source.modifiers:
                if modifier.type == 'SUBSURF':
                    subsurf_detected = True
                    subsurf_modifier = modifier  # Store the Subsurf modifier
                elif modifier.type == 'MULTIRES':
                    multires_detected = True
                    multires_modifier = modifier  # Store the Multires modifier
            
            if multires_detected and subsurf_detected:
                multires_view_level = multires_modifier.levels
                subdivision_view_level = subsurf_modifier.levels
                subsurf_modifier.levels = 0
                for level in range(multires_modifier.render_levels+1):
                    multires_modifier.levels = level
                    # Add a Surface deform modifier
                    modifier_name = "SurfaceDeformMULT" + str(level)
                    surfacedeform_modifier = copy_obj.modifiers.new(name=modifier_name, type='SURFACE_DEFORM')
                    surfacedeform_modifier.target = binding_source
                    # Manually invoke the bind operation
                    bpy.ops.object.surfacedeform_bind(modifier=modifier_name)
                    multires_modifier.levels = multires_modifier.render_levels
                subsurf_modifier.levels = subsurf_modifier.render_levels
                # Add a Surface deform modifier
                surfacedeform_modifier = copy_obj.modifiers.new(name="SurfaceDeformSUBD", type='SURFACE_DEFORM')
                surfacedeform_modifier.target = binding_source
                # Manually invoke the bind operation
                bpy.ops.object.surfacedeform_bind(modifier="SurfaceDeformSUBD")
                # Reset viewport levels
                multires_modifier.levels = multires_view_level
                subsurf_modifier.levels = subdivision_view_level
                subsurf_modifier.show_viewport = False
                
            if multires_detected and not subsurf_detected:
                # Save viewport level
                view_level = multires_modifier.levels
                for level in range(multires_modifier.render_levels+1):
                    multires_modifier.levels = level
                    # Add a Surface deform modifier
                    modifier_name = "SurfaceDeformMULT" + str(level)
                    surfacedeform_modifier = copy_obj.modifiers.new(name=modifier_name, type='SURFACE_DEFORM')
                    surfacedeform_modifier.target = binding_source
                    # Manually invoke the bind operation
                    bpy.ops.object.surfacedeform_bind(modifier=modifier_name)
                # Reset viewport level
                multires_modifier.levels = view_level
            
            if subsurf_detected and not multires_detected:
                # Save viewport level
                view_level = subsurf_modifier.levels
                for level in range(subsurf_modifier.render_levels+1):
                    subsurf_modifier.levels = level
                    # Add a Surface deform modifier
                    modifier_name = "SurfaceDeformSUBD" + str(level)
                    surfacedeform_modifier = copy_obj.modifiers.new(name=modifier_name, type='SURFACE_DEFORM')
                    surfacedeform_modifier.target = binding_source
                    # Manually invoke the bind operation
                    bpy.ops.object.surfacedeform_bind(modifier=modifier_name)
                # Reset viewport level
                subsurf_modifier.levels = view_level

            # Handle the case when neither modifier is detected
            if not subsurf_detected and not multires_detected:
                surfacedeform_modifier = copy_obj.modifiers.new(name="SurfaceDeform", type='SURFACE_DEFORM')
                surfacedeform_modifier.target = binding_source
                # Manually invoke the bind operation
                bpy.ops.object.surfacedeform_bind(modifier="SurfaceDeform")
            
            # Check if "FluidPainterGN" Geometry Nodes modifier exists in the original object
            original_gn_modifier = original_obj.modifiers.get("FluidPainterGN")
            if original_gn_modifier and context.scene.keep_fluid_gn:
                # Make the original object active and the copy object selected
                original_obj.select_set(True)
                copy_obj.select_set(True)
                bpy.context.view_layer.objects.active = original_obj
                
                # Copy the FluidPainterGN modifier from the original to the copy
                bpy.ops.object.modifier_copy_to_selected(modifier="FluidPainterGN")
                
                # Deselect original_obj
                original_obj.select_set(False)
                
                bpy.context.view_layer.objects.active = copy_obj
                
                # Set the "Is Mesh" flag to true
                copy_obj.modifiers["FluidPainterGN"]["Input_65"] = True
                            
                # Update the copy object
                copy_obj.update_tag()

            self.report({'INFO'}, "The Fluid" + copy_obj.name + " has been bound to" + binding_source.name)
                        
        return {'FINISHED'}


class MergeFluids(bpy.types.Operator):
    bl_idname = "object.merge_fluids"
    bl_label = "Merge Fluids"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
    
        if len(context.selected_objects) == 2:
            fluid_1 = context.active_object
            fluid_2 = bpy.context.selected_objects[0] if (fluid_1 != context.selected_objects[0]) else context.selected_objects[1]
            if is_fluid_curve_not_sel(fluid_1) and is_fluid_curve_not_sel(fluid_2) and fluid_1.name.split('.')[0] == fluid_2.name.split('.')[0]:
                return True
        else:
            return False

    def execute(self, context):
        
        fluid_1 = context.active_object
        fluid_2 = bpy.context.selected_objects[0] if (fluid_1 != context.selected_objects[0]) else context.selected_objects[1]

        bpy.ops.object.select_all(action='DESELECT')
        
        tools_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "Tools", "Fluid_Painter_Tools.blend")
    
        with bpy.data.libraries.load(tools_path, link=False) as (data_from, data_to):
            # Check if the selected_fluid object is in the current library
            if "Fluids Merge" in data_from.objects:
                # If found, assign it to data_to.objects and break the loop
                data_to.objects = ["Fluids Merge"]
        
        fluids_merge = None
        
        # Initialize a variable to hold the last appended object
        for obj in data_to.objects:
            if obj is not None:
                # Link the object to the scene
                bpy.context.scene.collection.objects.link(obj)
                # Assign the last appended object to the variable
                fluids_merge = obj
                
        if fluids_merge is not None:
            fluids_merge.select_set(True)
            bpy.context.view_layer.objects.active = fluids_merge
            print(fluids_merge.name)
            fluids_merge.name = fluid_1.name.split('.')[0] + " " + fluids_merge.name
            
        # Get the fluid collection
        fluids_collection = create_fluids_collection()
        
        # Unlink the object from its current collection (if it's linked to any)
        if fluids_merge.users_collection:
            current_collection = fluids_merge.users_collection[0]
            current_collection.objects.unlink(fluids_merge)
        
        fluids_collection.objects.link(fluids_merge)
        
        # Set the fluids_merge as the active object and select it
        bpy.context.view_layer.objects.active = fluids_merge
        fluids_merge.select_set(True)
        
        # Assign Fluids to Merge GN
        fluids_merge.modifiers["FluidPainterGN"]["Input_2"] = fluid_1
        fluids_merge.modifiers["FluidPainterGN"]["Input_3"] = fluid_2
        
        # Copy Drawing Objects and Collections
        if fluid_1.modifiers["FluidPainterGN"]["Input_41"]:
            fluids_merge.modifiers["FluidPainterGN"]["Input_7"] = fluid_1.modifiers["FluidPainterGN"]["Input_41"]
        if fluid_1.modifiers["FluidPainterGN"]["Input_36"]:
            fluids_merge.modifiers["FluidPainterGN"]["Input_8"] = fluid_1.modifiers["FluidPainterGN"]["Input_36"]
        if fluid_2.modifiers["FluidPainterGN"]["Input_41"]:
            fluids_merge.modifiers["FluidPainterGN"]["Input_9"] = fluid_2.modifiers["FluidPainterGN"]["Input_41"]
        if fluid_2.modifiers["FluidPainterGN"]["Input_36"]:
            fluids_merge.modifiers["FluidPainterGN"]["Input_10"] = fluid_2.modifiers["FluidPainterGN"]["Input_36"]
        
        # Copy GN Properties from the fluid 1
        modifier_fluid1 = fluid_1.modifiers.get('FluidPainterGN')
        modifier_fluid2 = fluid_2.modifiers.get('FluidPainterGN')
        modifier_fluidm = fluids_merge.modifiers.get('FluidPainterGN')
        
        if bpy.app.version >= (4, 0, 0):
            node_group_fluid1 = modifier_fluid1.node_group
            node_group_fluid2 = modifier_fluid2.node_group
            node_group_fluidm = modifier_fluidm.node_group
            inputs_fluid1 = node_group_fluid1.interface
            inputs_fluid2 = node_group_fluid2.interface
            inputs_fluidm = node_group_fluidm.interface

            # Copy other Properties
            for input_fluid1 in inputs_fluid1.items_tree:
                # Copy properties to fluid 2
                for input_fluid2 in inputs_fluid2.items_tree:
                    if input_fluid2.name == input_fluid1.name and input_fluid2.identifier != "Input_0" and input_fluid2.identifier != "Input_65" and input_fluid2.identifier != "Input_82" and input_fluid2.identifier != "Input_83" and input_fluid2.identifier != "Output_1" and input_fluid2.name.startswith('Drop') != True: 
                        if input_fluid2.in_out == "INPUT":
                            modifier_fluid2[input_fluid2.identifier] = modifier_fluid1[input_fluid1.identifier]
                        else:
                            modifier_fluid2[input_fluid2.identifier + "_attribute_name"] = modifier_fluid1[input_fluid1.identifier + "_attribute_name"]
                # Copy properties to fluid merge            
                for input_fluidm in inputs_fluidm.items_tree:
                    if input_fluidm.name == input_fluid1.name and input_fluidm.identifier != "Input_0" and input_fluidm.identifier != "Output_1": 
                        if input_fluidm.in_out == "INPUT":
                            modifier_fluidm[input_fluidm.identifier] = modifier_fluid1[input_fluid1.identifier]
                        else:
                            modifier_fluidm[input_fluidm.identifier + "_attribute_name"] = modifier_fluid1[input_fluid1.identifier + "_attribute_name"]
 
        else:
            node_group_fluid1 = modifier_fluid1.node_group
            node_group_fluid2 = modifier_fluid2.node_group
            node_group_fluidm = modifier_fluidm.node_group
            inputs_fluid1 = node_group_fluid1.inputs
            inputs_fluid2 = node_group_fluid2.inputs
            inputs_fluidm = node_group_fluidm.inputs
            outputs_fluid1 = node_group_fluid1.outputs
            outputs_fluid2 = node_group_fluid2.outputs
            outputs_fluidm = node_group_fluidm.outputs
        
            # Copy other Properties
            for input_fluid1 in inputs_fluid1:
                # Copy properties to fluid 2
                for input_fluid2 in inputs_fluid2:
                    if input_fluid2.name == input_fluid1.name and input_fluid2.identifier != "Input_0" and input_fluid2.identifier != "Input_65" and input_fluid2.identifier != "Input_82" and input_fluid2.identifier != "Input_83" and input_fluid2.name.startswith('Drop') != True:
                        modifier_fluid2[input_fluid2.identifier] = modifier_fluid1[input_fluid1.identifier]
                for input_fluidm in inputs_fluidm:
                    if input_fluidm.name == input_fluid1.name and input_fluidm.identifier != "Input_0":
                        modifier_fluidm[input_fluidm.identifier] = modifier_fluid1[input_fluid1.identifier]
                            
            for output_fluid1 in outputs_fluid1:
                # Copy properties to fluid merge  
                for output_fluid2 in outputs_fluid2:
                    if output_fluid2.name == output_fluid1.name and output_fluid1.identifier != "Output_1":
                        modifier_fluid2[output_fluid2.identifier + "_attribute_name"] = modifier_fluid1[output_fluid1.identifier + "_attribute_name"]
                for output_fluidm in outputs_fluidm:
                    if output_fluidm.name == output_fluid1.name and output_fluid1.identifier != "Output_1":
                        modifier_fluidm[output_fluidm.identifier + "_attribute_name"] = modifier_fluid1[output_fluid1.identifier + "_attribute_name"]


        # Set Fluids GN and hide
        fluids = [fluid_1, fluid_2]
        for fluid in fluids:
            fluid.modifiers["FluidPainterGN"]["Input_82"] = True
            fluid.hide_viewport = True
            fluid.hide_render = True
        
        self.report({'INFO'}, "Fluids Merged")
        
        return {'FINISHED'}
        

class UnmergeFluids(bpy.types.Operator):
    bl_idname = "object.unmerge_fluids"
    bl_label = "Merge Fluids"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        if (len(context.selected_objects) == 1 and
            context.active_object.modifiers.get('FluidPainterGN') is not None and
            context.active_object.modifiers['FluidPainterGN'].node_group.name.startswith("Fluid Merge") and
            context.active_object.modifiers['FluidPainterGN'].node_group is not None):
            return True 
        else:
            return False

    def execute(self, context):
    
        fluids_merge = context.selected_objects[0]
        modifier = context.active_object.modifiers.get('FluidPainterGN')
        fluid_1 = fluids_merge.modifiers["FluidPainterGN"]["Input_2"]
        fluid_2 = fluids_merge.modifiers["FluidPainterGN"]["Input_3"]
        
        # Set Fluids GN and unhide
        fluids = [fluid_1, fluid_2]
        for fluid in fluids:
            fluid.modifiers["FluidPainterGN"]["Input_82"] = False
            fluid.hide_viewport = False
            fluid.hide_render = False

        # Delete fluids_merge
        bpy.data.objects.remove(fluids_merge, do_unlink=True)
   
        return {'FINISHED'}


# Define the panel class for Fluid Editing
class FluidPainterToolsPanel:
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "Fluid Painter"


class FluidPainterToolsPanelMain(FluidPainterToolsPanel, bpy.types.Panel):
    bl_label = "Fluid Tools"
    bl_idname = "OBJECT_PT_fluid_tools"
    bl_options = {"DEFAULT_CLOSED"}
    
    def draw(self, context):
        layout = self.layout


class FluidPainterToolsPanelSetOrigin(FluidPainterToolsPanel, bpy.types.Panel):
    bl_parent_id = "OBJECT_PT_fluid_tools"
    bl_idname = "OBJECT_PT_fluid_tools_set_origin"
    bl_label = "Set Fluid Origin"
    bl_options = {"DEFAULT_CLOSED"}

    def draw(self, context):
    
        # Get the width of the panel region
        panel_width = context.region.width
    
        layout = self.layout
        
        layout.label(text="Set Fluid origin to its center")
        
        # Set Origin to Geometry
        layout.operator("object.set_origin", text="Set Origin to Geometry")


class FluidPainterToolsPanelConvertToMesh(FluidPainterToolsPanel, bpy.types.Panel):
    bl_parent_id = "OBJECT_PT_fluid_tools"
    bl_idname = "OBJECT_PT_fluid_tools_convert_to_mesh"
    bl_label = "Convert Fluid to Mesh"
    bl_options = {"DEFAULT_CLOSED"}

    def draw(self, context):
    
        # Get the width of the panel region
        panel_width = context.region.width
    
        layout = self.layout
        
        layout.label(text="Convert the Fluid to Mesh")
        
        # Convert to Mesh Operator
        layout.operator("object.convert_to_mesh", text="Convert Fluid to Mesh")


class FluidPainterToolsPanelConvertToFluid(FluidPainterToolsPanel, bpy.types.Panel):
    bl_parent_id = "OBJECT_PT_fluid_tools"
    bl_idname = "OBJECT_PT_fluid_tools_convert_to_fluid"
    bl_label = "Convert Curve to Fluid"
    bl_options = {"DEFAULT_CLOSED"}
    

    def draw(self, context):
        fluid_scale_factor = context.scene.fluid_scale
    
        # Get the width of the panel region
        panel_width = context.region.width
    
        layout = self.layout
        
        layout.label(text="Fluid Preset: "+ context.scene.fluid_painter.fluid)
        row = layout.row()
        row.label(text="Fluid Scale:")
        row.prop(context.scene, "fluid_scale", text="")
        
        # Convert to Mesh
        layout.operator("object.convert_to_fluid_operator", text="Convert Curve to Fluid")


class FluidPainterToolsPanelStickToMesh(FluidPainterToolsPanel, bpy.types.Panel):
    bl_parent_id = "OBJECT_PT_fluid_tools"
    bl_idname = "OBJECT_PT_fluid_tools_stick_to_mesh"
    bl_label = "Stick Fluid to Animated Object"
    bl_options = {"DEFAULT_CLOSED"}
    
    # Define the EnumProperty for stick methods
    stick_method_items = [
        ('METHOD_A', "Surface Deform", "Stick Fluid to object using surface deform"),
        ('METHOD_B', "Vertex Groups", "Stick Fluid to object transferring vertex groups"),
    ]

    bpy.types.Scene.stick_method = bpy.props.EnumProperty(
        items=stick_method_items,
        name="Stick Method",
        description="Choose the stick method",
        default='METHOD_A',  # Default method
    )

    def draw(self, context):
    
        # Get the width of the panel region
        panel_width = context.region.width
    
        layout = self.layout
        
         # Dropdown selector for stick methods
        layout.prop(context.scene, "stick_method", text="Method")
        
        # Input field for the source object (picker)
        layout.prop(context.scene, "sticking_mesh", text="Object")
        
        # Add a flag for keeping the fluid geometry node
        layout.prop(context.scene, "keep_fluid_gn", text="Keep Fluid Properties")
        
        # Add a label and a slider for spline resolution on the same line
        row = layout.row()
        row.label(text="Spline Resolution:")
        # Enable/disable the entire row based on the value of keep_fluid_gn
        row.enabled = context.scene.keep_fluid_gn
        row.prop(context.scene, "spline_resolution", text="")
        
        sticking_mesh = context.scene.sticking_mesh
            
        armature, current_position = get_armature_and_pose_position(sticking_mesh)
        
        if current_position:
            row = layout.row()
            row.label(text="Armature:")
            if current_position == 'REST':
                row.operator("object.switch_rest_pose", text="Pose Position")
            else:
                row.operator("object.switch_rest_pose", text="Rest Position") 
        
        if context.scene.stick_method == 'METHOD_A':
            # Button to add the surface deform modifier
            layout.operator("object.bind_to_mesh", text="Stick Fluid to Object")
            
            if sticking_mesh is not None:
                subsurf_detected = False
                multires_detected = False
                
                for modifier in sticking_mesh.modifiers:
                    if modifier.type == 'SUBSURF':
                        subsurf_detected = True
                    elif modifier.type == 'MULTIRES':
                        multires_detected = True
                        
                if subsurf_detected and not multires_detected:
                    wrapped_lines = wrap_label_text("Attention: Subdivision Modifier Detected on the Target Object. To prevent issues, one Surface Deform Modifier will be added to the Fluid for each Subd Level.", panel_width)
                    box = layout.box()
                    for line in wrapped_lines:
                        box.label(text=line)
                elif multires_detected and not subsurf_detected:
                    wrapped_lines = wrap_label_text("Attention: Multires Modifier Detected on the Target Object. To prevent issues, one Surface Deform Modifier will be added to the Fluid for each Subd Level.", panel_width)
                    box = layout.box()
                    for line in wrapped_lines:
                        box.label(text=line)
                elif subsurf_detected and multires_detected:
                    wrapped_lines = wrap_label_text("Attention: Both Subdivision and Multires Modifiers Detected on the Target Object. One Surface Deform Modifier will be added to the Fluid for each Subdivision Level of the Multires Modifier, and one for the Render Level of the Subdivision Modifier. After sticking, the Subdivision Modifier will be disabled in viewport to avoid issues.", panel_width)
                    box = layout.box()
                    for line in wrapped_lines:
                        box.label(text=line)              
                                            
        elif context.scene.stick_method == 'METHOD_B':

            # Button to add the Data Transfer modifier and Armature modifier
            layout.operator("object.vertex_groups_transfer", text="Stick Fluid to Object")
            
            if not armature:
                wrapped_lines = wrap_label_text("Warning! Mesh must have an armature", panel_width)
                box = layout.box()
                for line in wrapped_lines:
                    box.label(text=line)
        
        # Check if the collection containing the stiching mesh is hidden
        if sticking_mesh is not None:
            if sticking_mesh.users_collection:
                collection = sticking_mesh.users_collection[0]  # Assuming the mesh is only in one collection
                view_layer = context.view_layer
                for layer_collection in view_layer.layer_collection.children:
                    if layer_collection.collection == collection:
                        if layer_collection.exclude:
                            wrapped_lines = wrap_label_text("Warning! Target Mesh Collection is excluded from the viewport", panel_width)
                            box = layout.box()
                            for line in wrapped_lines:
                                box.label(text=line)
                    

class FluidPainterToolsPanelMergeFluids(FluidPainterToolsPanel, bpy.types.Panel):
    bl_parent_id = "OBJECT_PT_fluid_tools"
    bl_idname = "OBJECT_PT_fluid_merge"
    bl_label = "Merge Fluids"
    bl_options = {"DEFAULT_CLOSED"}

    def draw(self, context):
        # Get the width of the panel region
        panel_width = context.region.width
        
        fluid_1 = None
        fluid_2 = None
        
        layout = self.layout
        
        if len(context.selected_objects) == 2:
            fluid_1 = context.active_object
            fluid_2 = bpy.context.selected_objects[0] if (fluid_1 != context.selected_objects[0]) else context.selected_objects[1]
            
            # Check if the two selected objects are fluids
            if is_fluid_curve_not_sel(fluid_1) and is_fluid_curve_not_sel(fluid_2):
                # Check if fluids are of the same type
                if fluid_1.name.split('.')[0] == fluid_2.name.split('.')[0]:
                    layout.label(text="Fluid 1: " + fluid_1.name)
                    layout.label(text="Fluid 2: " + fluid_2.name + "*")
                    wrapped_lines = wrap_label_text("* Fluid 2 will inherit properties from Fluid 1", panel_width)
                    box = layout.box()
                    for line in wrapped_lines:
                        box.label(text=line)
                else:
                    layout.label(text="The Two fluids must be of the same type")
            
            else:
                layout.label(text="The objects must be of fluid type")
            
        elif len(context.selected_objects) == 1 and context.active_object.modifiers.get('FluidPainterGN') is not None and context.active_object.modifiers['FluidPainterGN'].node_group.name.startswith("Fluid Merge") == True:
            layout.label(text="Merged Fluid: " + context.active_object.name)
        else:
            layout.label(text="Select two fluids to merge")
        
        row = layout.row()
        col = row.column()
        col.operator("object.merge_fluids", text='Merge')
        col.operator("object.unmerge_fluids", text='Unmerge')


# Define a property for spline resolution
bpy.types.Scene.fluid_scale = bpy.props.FloatProperty(
    name="Fluid Scale",
    description="Scale of the Fluid",
    default=1.0,
    min=0.1,
    max=100000.0,
)

# Update the Scene property for selecting the source object
bpy.types.Scene.sticking_mesh = bpy.props.PointerProperty(
    type=bpy.types.Object,
    name="Sticking Mesh",
    description="Select the mesh you want the fluid to stick to",
)

# Define a property for keeping the fluid geometry node
bpy.types.Scene.keep_fluid_gn = bpy.props.BoolProperty(
    name="Keep fluid geometry node",
    description="Flag to indicate whether to keep the fluid geometry node",
    default=False,
)

# Define a property for spline resolution
bpy.types.Scene.spline_resolution = bpy.props.IntProperty(
    name="Spline Resolution",
    description="Resolution for the spline",
    default=12,
    min=2,
    soft_max=64,
)