import bpy
from . import main_fluid_version
from .FP_Append import library_file_path
from .FP_Utilities import is_fluid_curve, is_fluid_curve_uptodate, debug_print, wrap_label_text, clear_tags, get_additional_library_file_path, compare_versions, set_frame_driver
from mathutils import Color


# Operator to update the selected curve
class UpdateCurveOperator(bpy.types.Operator):
    bl_idname = "object.update_curve_operator"
    bl_label = "Update Curve"
    bl_options = {'REGISTER', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        # Check if there are any selected objects
        if not context.selected_objects:
            return False
    
        # Check if there's just one selected object
        if context.mode != 'OBJECT' and len(context.selected_objects) != 1:
            return False
            
        # Get the selected object
        selected_object = context.selected_objects[0]
        
        # Check if the selected object is a curve
        if selected_object.type != 'CURVE':
            return False

        # Check if the selected object has a geometry node modifier with a node group name that starts with "fluid painter"
        has_fluid_painter_modifier = any(mod.type == 'NODES' and mod.node_group.name.startswith("Fluid Painter") for mod in selected_object.modifiers)
        
        # Return True only if the conditions are met
        if not has_fluid_painter_modifier:
            return False
        
        return not is_fluid_curve(selected_object) or not is_fluid_curve_uptodate(selected_object)
        
        debug_print(f"Has Fluid Painter Modifier? {has_fluid_painter_modifier}")
        debug_print(f"Is Fluid Curve? {is_fluid_curve(selected_object)}")
        debug_print(f"Is Fluid Curve up to date? {is_fluid_curve_uptodate(selected_object)}")

            
    def execute(self, context):
        update_materials = context.scene.update_materials
    
        # Get the active object (selected curve)
        selected_object = bpy.context.selected_objects[0]
        smooth_iteration = None
              
        # Drop Check Logic
        set_drop_size = False
        # Check if the object has the "Fluid_Ver" custom property
        if "fluid_ver" not in selected_object:
            set_drop_size = True
        elif compare_versions(selected_object["fluid_ver"], "1.2.15", digits_to_check=None) <0:
            set_drop_size = True
            
        clear_tags(selected_object)
        
        # Add version property
        selected_object["fluid_ver"] = main_fluid_version
    
        # Add "fluid_init" property and set to true
        selected_object["fluid_init"] = 1

        # Check if the curve name contains a "+"
        if "+" in selected_object.name:
            # Add "fluid_plus" property and set to true
            selected_object["fluid_plus"] = 1
    
        # Save smooth iterations, remove smooth modifier and Update GN Inputs, 
        if selected_object.modifiers:
            for modifier in selected_object.modifiers:
                if modifier.type == 'SMOOTH':
                    smooth_iteration = modifier.iterations
                    debug_print(f"Smooth Iteration: {smooth_iteration}")
                    selected_object.modifiers.remove(modifier)

            for modifier in selected_object.modifiers:
                if modifier.type == 'NODES':
                    # Check if the Geometry Node modifier has a node group with the name "Fluid Painter"
                    if modifier.node_group and modifier.node_group.name.startswith("Fluid Painter"):
                    
                        # Rename the Geometry Node modifier to "FluidPainterGN"
                        modifier.name = "FluidPainterGN"
                        
                        # Import the node group from the specified .blend file
                        with bpy.data.libraries.load(library_file_path, link=False) as (data_from, data_to):
                            # Append the node group from the source file
                            data_to.node_groups = [name for name in data_from.node_groups if name == "Fluid Painter"]
                            
                        # Get the newly appended 'Fluid Painter' (it will retain its original name)
                        imported_node_group = data_to.node_groups[0]

                        # Assign the imported node group to the modifier
                        modifier.node_group = imported_node_group
                        
                        # Update the object
                        selected_object.update_tag()
                        
                        if set_drop_size:
                            modifier["Input_34"] += 1
                        
                        if smooth_iteration:
                            debug_print(f"Smooth Iteration set: {smooth_iteration}")
                            modifier["Input_64"] = min(round(smooth_iteration / 10), 5)
                        
                        # Set End Frame
                        set_frame_driver(modifier, "Input_77", "frame_start")
                        set_frame_driver(modifier, "Input_76", "frame_end")
                        
                        # Import the material only if update_materials is True
                        if update_materials:
                            # Define the list of input names
                            input_names = ["Input_3", "Input_25", "Input_26"]

                            # Define the list of library paths to check
                            library_paths = [library_file_path]

                            additional_library_file_path = get_additional_library_file_path()

                            # Check if additional_library_file_path is not empty and add it to the list
                            if additional_library_file_path:
                                library_paths.append(additional_library_file_path)
                                print(library_paths)

                            for input_name in input_names:
                                input_material = modifier[input_name]

                                if input_material:
                                    # Get the material name before the dot (if any)
                                    material_name = input_material.name.split(".")[0]

                                    # Check if the input_name is "Input_25" or "Input_26"
                                    if input_name in ["Input_25", "Input_26"]:
                                        # Check if the word "layer" is not in the material name
                                        if "Layer" not in material_name:
                                            material_name = material_name.replace("Fluid", "Fluid Layer")

                                    material_found = False

                                    # Iterate through the library paths and try to import the material
                                    for library_path in library_paths:
                                        with bpy.data.libraries.load(library_path, link=False) as (data_from, data_to):
                                            if material_name in data_from.materials:
                                                data_to.materials = [material_name]
                                                material_found = True
                                                break
                                    
                                    if material_found:
                                        imported_material = data_to.materials[0]
                                        
                                        # Assign the material to the corresponding input in the Geometry Node modifier
                                        modifier[input_name] = imported_material
                                        
                                        # Check if the material (without split) is in the object materials slot
                                        for slot in selected_object.material_slots:
                                            if slot.material and slot.material.name == input_material.name:
                                                # Replace the existing material with the imported_material
                                                slot.material = imported_material
                                                break
                                                
                                        node_tree = imported_material.node_tree
                                        
                                        # Find the Principled BSDF node
                                        #principled_node = None
                                        # node in node_tree.nodes:
                                            #if node.type == 'BSDF_PRINCIPLED':
                                                #principled_node = node
                                                #print("Principled Node Found")
                                                #break

                                        # Traverse the node tree to find nodes that contain attributes
                                        for node in node_tree.nodes:
                                            if node.type == 'ATTRIBUTE':
                                                print("Attribute Name:", node.attribute_name)
                                                if node.attribute_name == "distance":
                                                    modifier["Output_39_attribute_name"] = "distance"
                                                elif node.attribute_name == "base_color":
                                                    modifier["Output_73_attribute_name"] = "base_color"
                                                    #if principled_node:
                                                        #base_color = principled_node.inputs["Base Color"].default_value
                                                        #modifier["Input_74"] = base_color
                                                elif node.attribute_name == "shadow":
                                                    modifier["Output_84_attribute_name"] = "shadow"
                                                elif node.attribute_name == "uv_scale":
                                                    modifier["Output_52_attribute_name"] = "uv_scale"
                                                elif node.attribute_name == "scale":
                                                    modifier["Output_62_attribute_name"] = "scale"
                                                elif node.attribute_name == "pos":
                                                    modifier["Output_69_attribute_name"] = "pos"
                                                elif node.attribute_name == "rot":
                                                    modifier["Output_71_attribute_name"] = "rot"
                                                elif node.attribute_name == "bbox":
                                                    modifier["Output_72_attribute_name"] = "bbox"
                                                elif node.attribute_name == "layer_a_random":
                                                    modifier["Output_66_attribute_name"] = "layer_a_random"
                                                elif node.attribute_name == "layer_b_random":
                                                    modifier["Output_67_attribute_name"] = "layer_b_random"   
                                    else:
                                        print(f"Material '{material_name}' not found in the libraries.")
        return {'FINISHED'}


# Panel to display the update button
class FluidPainterUpdatePanel(bpy.types.Panel):
    bl_label = "Fluid Update"
    bl_idname = "OBJECT_PT_fluid_update"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "Fluid Painter"
    bl_options = {"DEFAULT_CLOSED"}
  
    def draw(self, context):
    
        # Get the width of the panel region
        panel_width = context.region.width
    
        selected_object = bpy.context.active_object
        layout = self.layout
        
        # Display the Update Curve button
        layout.operator("object.update_curve_operator", text="Update Fluid to the latest version")
        
        # Add toggle buttons for update_materials
        layout.prop(context.scene, "update_materials", text="Update Materials")


# Define a scene property for updating materials
bpy.types.Scene.update_materials = bpy.props.BoolProperty(
    name="Update Materials",
    description="Toggle to update materials",
    default=True
)