import bpy
import os
import textwrap
from . import main_fluid_version


# Initialize additional_library_file_path to None
additional_library_file_path = None


# Set addon path
def set_addon_path(other_addon_name, library_folder, library_file):
    global additional_library_file_path

    addon_module = __import__(other_addon_name)
    additional_script_directory = os.path.dirname(os.path.realpath(addon_module.__file__))
    additional_library_file_path = os.path.join(additional_script_directory, library_folder, library_file)
    print("addon path defined for Update")
    

def get_additional_library_file_path():
    return additional_library_file_path    


def wrap_label_text(text, width):
    char_width = 7
    wrapped_lines = textwrap.wrap(text, width=width // char_width)
    return wrapped_lines

      
def compare_versions(version1, version2, digits_to_check=None):
    """
    Compare two version strings in "x.y.z" format, considering a specified number of digits.
    Returns:
    -1 if version1 < version2
     0 if version1 == version2
     1 if version1 > version2
    """

    def parse_version(v, digits_to_check):
        components = [int(x) for x in v.split(".")]
        if digits_to_check is not None:
            components = components[:digits_to_check]
        return components

    v1 = parse_version(version1, digits_to_check)
    v2 = parse_version(version2, digits_to_check)

    if v1 < v2:
        return -1
    elif v1 == v2:
        return 0
    else:
        return 1

        
def debug_print(*args):
    if bpy.app.debug:
        message = ' '.join(map(str, args))
        print("[DEBUG]", message)
    
    
def is_fluid_curve(obj):
    # Check if the curve is a valid fluid
    if obj is not None:
        return (
            
            (obj.type == 'CURVE' or obj.type == "MESH")
            and "fluid_init" in obj
            and obj.get("fluid_init")
            and obj.select_get()
            and any(mod.type == 'NODES' and mod.name == "FluidPainterGN" and mod.node_group.name.startswith("Fluid Painter") for mod in obj.modifiers)
        )
    else:
        return False


def is_fluid_curve_not_sel(obj):
    # Check if the curve is a valid fluid
    if obj is not None:
        return (
            
            (obj.type == 'CURVE' or obj.type == "MESH")
            and "fluid_init" in obj
            and obj.get("fluid_init")
            #and obj.select_get()
            and any(mod.type == 'NODES' and mod.name == "FluidPainterGN" and mod.node_group.name.startswith("Fluid Painter") for mod in obj.modifiers)
        )
    else:
        return False

    
def is_fluid_curve_uptodate(obj):
    # Check if "fluid_ver" property exists and is greater than or equal to the current version
    curr_fluid_version = obj.get("fluid_ver")
        
    # Check if both version checks are true
    return ( 
        curr_fluid_version is not None and compare_versions(curr_fluid_version, main_fluid_version, digits_to_check=None) == 0
    )


def are_selected_meshes_valid():
    selected_objects = bpy.context.selected_objects
    return selected_objects and all(
        (obj.type == 'MESH') or (obj.type == 'CURVE' and (obj.get("fluid_init", False) or "fluid_init" not in obj))
        for obj in selected_objects
    )

    
def set_drawing_objects(fluid, selected_objects):
    # Remove the fluid object from the selected objects list, if it exists
    selected_objects = [obj for obj in selected_objects if obj != fluid]
    # Define selected_mesh as a variable
    selected_mesh = None
    # Define new_collection as a variable
    new_collection = None

    # Reset Drawing Objects
    if fluid.modifiers["FluidPainterGN"]["Input_41"] is not None:
        fluid.modifiers["FluidPainterGN"]["Input_41"] = None
    if fluid.modifiers["FluidPainterGN"]["Input_36"] is not None:
        collection_to_delete = fluid.modifiers["FluidPainterGN"]["Input_36"]
        bpy.data.collections.remove(collection_to_delete)
        fluid.modifiers["FluidPainterGN"]["Input_36"] = None

    # Check if there is just one selected mesh and/or fluid curve objects
    if len(selected_objects) == 1 and (
        selected_objects[0].type == 'MESH'
        or (
            selected_objects[0].type == 'CURVE'
            and (selected_objects[0].get("fluid_init", False) or "fluid_init" not in selected_objects[0])
        )
    ):
        selected_mesh = selected_objects[0]

    # Check if there are multiple selected mesh and/or fluid curve objects
    elif len(selected_objects) >= 2 and all(
        (obj.type == 'MESH' or (obj.type == 'CURVE' and (obj.get("fluid_init", False) or "fluid_init" not in obj)))
        for obj in selected_objects
    ):
        # Create a new collection
        new_collection_name = fluid.name + " Collection"
        new_collection = bpy.data.collections.new(new_collection_name)

        # Check if the "Fluids" collection exists
        fluids_collection = bpy.data.collections.get("Fluids")
        if fluids_collection is not None:
            fluids_collection.children.link(new_collection)
        else:
            # Create the "Fluids" collection and link the new collection
            fluids_collection = bpy.data.collections.new("Fluids")
            bpy.context.scene.collection.children.link(fluids_collection)
            fluids_collection.children.link(new_collection)

        # Add selected objects to the new collection
        for obj in selected_objects:
            new_collection.objects.link(obj)

    if selected_mesh is not None:
        fluid.modifiers["FluidPainterGN"]["Input_41"] = selected_mesh
    elif new_collection is not None:
        fluid.modifiers["FluidPainterGN"]["Input_36"] = new_collection
    

def create_fluids_collection():
    # Check if "Fluids" collection exists in the scene
    scene = bpy.context.scene
    fluids_collection = None
    
    for collection in scene.collection.children:
        if collection.name.startswith("Fluid"):
            fluids_collection = collection
            break
    
    if fluids_collection is None:
        fluids_collection = bpy.data.collections.new("Fluids")
        scene.collection.children.link(fluids_collection)
    
    # Check if "Fluids" collection is excluded from view layer
    view_layer = bpy.context.view_layer
    for collection in view_layer.layer_collection.children:
        if collection.collection == fluids_collection:
            if collection.exclude:
                # Include the "Fluids" collection
                collection.exclude = False
            break
    
    # Set the "Fluids" collection as the active collection
    bpy.context.view_layer.active_layer_collection = bpy.context.view_layer.layer_collection.children[fluids_collection.name]
    
    return fluids_collection
    
    
def clear_tags(obj):
    if obj is not None:
        if 'fluid_init' in obj:
            del obj['fluid_init']
        if 'fluid_ver' in obj:
            del obj['fluid_ver']
        if 'fluid_blender_ver' in obj:
            del obj['fluid_blender_ver']
        if 'fluid_plus' in obj:
            del obj['fluid_plus']
            

def remove_frame_driver(modifier, input_name):
    if modifier:
        # Construct the full data path of the driver
        driver_data_path = '["{}"]'.format(input_name)
        
        # Remove the driver
        modifier.driver_remove(driver_data_path)
        print("Driver removed.")
    else:
        print("Modifier not found.")


def set_frame_driver(modifier, input_name, data_path):
    remove_frame_driver(modifier, input_name)
    
    # Add the new driver
    driver = modifier.driver_add('["{}"]'.format(input_name)).driver
    driver.type = 'AVERAGE'
    var = driver.variables.new()
    var.type = 'SINGLE_PROP'
    var.targets[0].id_type = 'SCENE'
    var.targets[0].id = bpy.context.scene
    var.targets[0].data_path = data_path
    driver.expression = 'var'
